#
# ur.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .utils import is_ur_type


class InvalidType(Exception):
    pass


class UR:
    def __init__(self, type, cbor):
        if not is_ur_type(type):
            raise InvalidType()

        self.type = type
        self.cbor = cbor

    def __eq__(self, obj):
        if obj == None:
            return False
        return self.type == obj.type and self.cbor == obj.cbor



#
# (c) Copyright 2023 by Coinkite Inc. This file is in the public domain.
#
# - joins QR codes
#
from .consts import KNOWN_FILETYPES
from .utils import decode_data


def join_qrs(parts):
    # take a bunch of scanned data.
    # - put into order, decode, return type code and raw data bytes
    # - lazy desktop code here
    hdr = set(p[0:6] for p in parts)
    assert len(hdr) == 1, "conflicting/variable filetype/encodings/sizes"
    hdr = hdr.pop()

    assert hdr[0:2] == "B$", "fixed header not found, expected B$"
    encoding = hdr[2]
    file_type = hdr[3]
    num_parts = int(hdr[4:6], 36)

    assert num_parts >= 1, "zero parts?"
    assert encoding in "H2Z", f"bad encoding: {encoding}"
    assert file_type in KNOWN_FILETYPES, f"bad file type: {encoding}"

    # ok to have dups here, just need them all
    data = {}
    for p in parts:
        idx = int(p[6:8], 36)
        assert idx < num_parts, f"got part {idx} but only expecting {num_parts}"

        if idx in data:
            assert data[idx] == p[8:], f"dup part 0x{idx:02x} has wrong content"
        else:
            data[idx] = p[8:]

    missing = set(range(num_parts)) - set(data)
    assert not missing, f"parts missing: {missing!r}"

    parts = [data[i] for i in range(num_parts)]

    raw = decode_data(parts, encoding)

    # maybe: decode objects here... U=>text, C=>obj, J=>obj

    return file_type, raw


# EOF



#
# utils.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .crc32 import crc32, crc32n


def crc32_bytes(buf):
    checksum = crc32n(buf)
    return checksum


def crc32_int(buf):
    return crc32(buf)


def data_to_hex(buf):
    return "".join("{:02x}".format(x) for x in buf)


def int_to_bytes(n):
    # return n.to_bytes((n.bit_length() + 7) // 8, 'big')
    return n.to_bytes(4, "big")


def bytes_to_int(buf):
    return int.from_bytes(buf, "big")


def string_to_bytes(s):
    return bytes(s, "utf8")


def is_ur_type(ch):
    if "a" <= ch and ch <= "z":
        return True
    if "0" <= ch and ch <= "9":
        return True
    if ch == "-":
        return True
    return False


def partition(s, n):
    return [s[i : i + n] for i in range(0, len(s), n)]


# Split the given sequence into two parts returned in a tuple
# The first entry in the tuple has the first `count` values.
# The second entry in the tuple has the remaining values.
def split(buf, count):
    return (buf[0:count], buf[count:])


def join_lists(lists):
    # return [y for x in lists for y in x]
    return sum(lists, [])


def join_bytes(list_of_ba):
    out = bytearray()
    for ba in list_of_ba:
        out.extend(ba)
    return out


def xor_into(target, source):
    count = len(target)
    assert count == len(source)  # Must be the same length
    for i in range(count):
        target[i] ^= source[i]


def xor_with(a, b):
    target = a
    xor_into(target, b)
    return target


def take_first(s, count):
    return s[0:count]


def drop_first(s, count):
    return s[count:]



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun
# Copyright (c) 2015 Sokolov Yura
# Copyright (c) 2013 Fritz Grimpen

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# coding: utf-8


class Tagging(object):
    __slots__ = ("tag", "obj")

    def __init__(self, tag, obj):
        self.tag = tag
        self.obj = obj

    def __eq__(self, other):
        return isinstance(other, Tagging) and self.tag == other.tag and self.obj == other.obj


class Mapping(object):
    __slots__ = "map"

    def __init__(self, map):
        self.map = map

    def mapping(obj):
        return Mapping(obj)


class DataItem(Tagging):
    def __init__(self, tag, map):
        super().__init__(tag, Mapping(map))
        self.tag = tag
        self.map = map


class _Undefined(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance

    def __str__(self):
        return "Undefined"

    def __repr__(self):
        return "Undefined"


Undefined = _Undefined()

__all__ = ["Tagging", "Mapping", "DataItem"]



#
# xoshiro256.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

import sys

try:
    import uhashlib as hashlib
except:
    try:
        import hashlib
    except:
        sys.exit("ERROR: No hashlib or uhashlib implementation found (required for sha256)")

from ..ur.constants import MAX_UINT64
from ..ur.utils import int_to_bytes, string_to_bytes

# Original Info:
# Written in 2018 by David Blackman and Sebastiano Vigna (vigna@acm.org)

# To the extent possible under law, the author has dedicated all copyright
# and related and neighboring rights to this software to the public domain
# worldwide. This software is distributed without any warranty.

# See <http://creativecommons.org/publicdomain/zero/1.0/>.

# This is xoshiro256** 1.0, one of our all-purpose, rock-solid
# generators. It has excellent (sub-ns) speed, a state (256 bits) that is
# large enough for any parallel application, and it passes all tests we
# are aware of.

# For generating just floating-point numbers, xoshiro256+ is even faster.

# The state must be seeded so that it is not everywhere zero. If you have
# a 64-bit seed, we suggest to seed a splitmix64 generator and use its
# output to fill s.


def rotl(x, k):
    return ((x << k) | (x >> (64 - k))) & MAX_UINT64


JUMP = [0x180EC6D33CFD0ABA, 0xD5A61266F0C9392C, 0xA9582618E03FC9AA, 0x39ABDC4529B1661C]
LONG_JUMP = [
    0x76E15D3EFEFDCBBF,
    0xC5004E441C522FB3,
    0x77710069854EE241,
    0x39109BB02ACBE635,
]


class Xoshiro256:
    def __init__(self, arr=None):
        self.s = [0] * 4
        if arr != None:
            self.s[0] = arr[0]
            self.s[1] = arr[1]
            self.s[2] = arr[2]
            self.s[3] = arr[3]

    def _set_s(self, arr):
        for i in range(4):
            o = i * 8
            v = 0
            for n in range(8):
                v <<= 8
                v |= arr[o + n]
            self.s[i] = v

    def _hash_then_set_s(self, buf):
        m = hashlib.sha256()
        m.update(buf)
        digest = m.digest()
        self._set_s(digest)

    @classmethod
    def from_int8_array(cls, arr):
        x = Xoshiro256()
        x._set_s(arr)
        return x

    @classmethod
    def from_bytes(cls, buf):
        x = Xoshiro256()
        x._hash_then_set_s(buf)
        return x

    @classmethod
    def from_crc32(cls, crc32):
        x = Xoshiro256()
        buf = int_to_bytes(crc32)
        x._hash_then_set_s(buf)
        return x

    @classmethod
    def from_string(cls, s):
        x = Xoshiro256()
        buf = string_to_bytes(s)
        x._hash_then_set_s(buf)
        return x

    def next(self):
        result = (rotl((self.s[1] * 5) & MAX_UINT64, 7) * 9) & MAX_UINT64
        t = (self.s[1] << 17) & MAX_UINT64

        self.s[2] ^= self.s[0]
        self.s[3] ^= self.s[1]
        self.s[1] ^= self.s[2]
        self.s[0] ^= self.s[3]

        self.s[2] ^= t

        self.s[3] = rotl(self.s[3], 45) & MAX_UINT64

        return result

    def next_double(self):
        m = float(MAX_UINT64) + 1
        nxt = self.next()
        return nxt / m

    def next_int(self, low, high):
        return int(self.next_double() * (high - low + 1) + low) & MAX_UINT64

    def next_byte(self):
        return self.next_int(0, 255)

    def next_data(self, count):
        result = bytearray()
        for i in range(count):
            result.append(self.next_byte())
        return result

    def jump(self):
        global JUMP

        s0 = 0
        s1 = 0
        s2 = 0
        s3 = 0
        for i in range(len(JUMP)):
            for b in range(64):
                if JUMP[i] & (1 << b):
                    s0 ^= self.s[0]
                    s1 ^= self.s[1]
                    s2 ^= self.s[2]
                    s3 ^= self.s[3]
                self.next()

        self.s[0] = s0
        self.s[1] = s1
        self.s[2] = s2
        self.s[3] = s3

    def long_jump(self):
        global LONG_JUMP

        s0 = 0
        s1 = 0
        s2 = 0
        s3 = 0
        for i in range(len(LONG_JUMP)):
            for b in range(64):
                if LONG_JUMP[i] & (1 << b):
                    s0 ^= self.s[0]
                    s1 ^= self.s[1]
                    s2 ^= self.s[2]
                    s3 ^= self.s[3]
                self.next()

        self.s[0] = s0
        self.s[1] = s1
        self.s[2] = s2
        self.s[3] = s3



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import io

from ..urtypes.cbor import DataItem, decoder, encoder


class RegistryType:
    def __init__(self, type, tag):
        self.type = type
        self.tag = tag


class RegistryItem:
    @classmethod
    def registry_type(cls):
        raise NotImplementedError()

    @classmethod
    def mapping(cls, item):
        if isinstance(item, DataItem):
            registry_type = cls.registry_type()
            if (registry_type is None and item.tag is None) or (
                registry_type is not None and registry_type.tag == item.tag
            ):
                return item.map
        return item

    @classmethod
    def from_data_item(cls, item):
        raise NotImplementedError()

    def to_data_item(self):
        raise NotImplementedError()

    @classmethod
    def from_cbor(cls, cbor_payload):
        cbor_decoder = decoder.Decoder(io.BytesIO(cbor_payload))
        return cls.from_data_item(cbor_decoder.decode())

    def to_cbor(self):
        cbor_encoder = encoder.Encoder(io.BytesIO())
        cbor_encoder.encode(self.to_data_item())
        v = cbor_encoder.output.getvalue()
        cbor_encoder.output.close()
        return bytearray(v)



#
# fountain_encoder.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

import math

from .cbor_lite import CBORDecoder, CBOREncoder
from .constants import MAX_UINT32, MAX_UINT64
from .fountain_utils import choose_fragments
from .utils import crc32_int, data_to_hex, split, xor_into


class InvalidHeader(Exception):
    pass


class Part:
    def __init__(self, seq_num, seq_len, message_len, checksum, data):
        self.seq_num = seq_num
        self.seq_len = seq_len
        self.message_len = message_len
        self.checksum = checksum
        self.data = data

    @staticmethod
    def from_cbor(cbor_buf):
        try:
            decoder = CBORDecoder(cbor_buf)
            (array_size, _) = decoder.decodeArraySize()
            if array_size != 5:
                raise InvalidHeader()

            (seq_num, _) = decoder.decodeUnsigned()
            if seq_num > MAX_UINT64:  # TODO: Do something better with this check
                raise InvalidHeader()

            (seq_len, _) = decoder.decodeUnsigned()
            if seq_len > MAX_UINT64:
                raise InvalidHeader()

            (message_len, _) = decoder.decodeUnsigned()
            if message_len > MAX_UINT64:
                raise InvalidHeader()

            (checksum, _) = decoder.decodeUnsigned()
            if checksum > MAX_UINT64:
                raise InvalidHeader()

            (data, _) = decoder.decodeBytes()

            return Part(seq_num, seq_len, message_len, checksum, data)
        except Exception:
            raise InvalidHeader()

    def cbor(self):
        encoder = CBOREncoder()
        encoder.encodeArraySize(5)
        encoder.encodeInteger(self.seq_num)
        encoder.encodeInteger(self.seq_len)
        encoder.encodeInteger(self.message_len)
        encoder.encodeInteger(self.checksum)
        encoder.encodeBytes(self.data)
        return encoder.get_bytes()

    def seq_num(self):
        return self.seq_num

    def seq_len(self):
        return self.seq_len

    def message_len(self):
        return self.message_len

    def checksum(self):
        return self.checksum

    def data(self):
        return self.data

    def description(self):
        return "seqNum:{}, seqLen:{}, messageLen:{}, checksum:{}, data:{}".format(
            self.seq_num,
            self.seq_len,
            self.message_len,
            self.checksum,
            data_to_hex(self.data),
        )


class FountainEncoder:
    def __init__(self, message, max_fragment_len, first_seq_num=0, min_fragment_len=10):
        assert len(message) <= MAX_UINT32
        self.message_len = len(message)
        self.checksum = crc32_int(message)
        self.fragment_len = FountainEncoder.find_nominal_fragment_length(
            self.message_len, min_fragment_len, max_fragment_len
        )
        self.fragments = FountainEncoder.partition_message(message, self.fragment_len)
        self.seq_num = first_seq_num

    @staticmethod
    def find_nominal_fragment_length(message_len, min_fragment_len, max_fragment_len):
        assert message_len > 0
        assert min_fragment_len > 0
        assert max_fragment_len >= min_fragment_len
        max_fragment_count = message_len // min_fragment_len
        fragment_len = None

        for fragment_count in range(1, max_fragment_count + 1):
            fragment_len = math.ceil(message_len / fragment_count)
            if fragment_len <= max_fragment_len:
                break

        assert fragment_len != None
        return fragment_len

    @staticmethod
    def partition_message(message, fragment_len):
        remaining = message
        fragments = []
        while len(remaining) != 0:
            (fragment, remaining) = split(remaining, fragment_len)
            padding = fragment_len - len(fragment)
            while padding > 0:
                fragment.append(0)
                padding -= 1
            fragments.append(fragment)

        return fragments

    def last_part_indexes(self):
        return self.last_part_indexes

    def seq_len(self):
        return len(self.fragments)

    # This becomes `true` when the minimum number of parts
    # to relay the complete message have been generated
    def is_complete(self):
        return self.seq_num >= self.seq_len()

    # True if only a single part will be generated.
    def is_single_part(self):
        return self.seq_len() == 1

    def next_part(self):
        self.seq_num += 1
        self.seq_num = self.seq_num % MAX_UINT32  # wrap at period 2^32
        indexes = choose_fragments(self.seq_num, self.seq_len(), self.checksum)
        mixed = self.mix(indexes)
        data = bytes(mixed)
        return Part(self.seq_num, self.seq_len(), self.message_len, self.checksum, data)

    def mix(self, indexes):
        result = [0] * self.fragment_len
        for index in indexes:
            xor_into(result, self.fragments[index])
        return result



import base64
import logging
import re
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Set, Tuple

import bdkpython as bdk

from bitcoin_qr_tools import bbqr
from bitcoin_qr_tools.bbqr.consts import FILETYPE_NAMES, KNOWN_FILETYPES
from bitcoin_qr_tools.data import Data

from .ur.ur_decoder import URDecoder
from .urtypes.bytes import Bytes as UR_BYTES
from .urtypes.crypto import PSBT as UR_PSBT
from .urtypes.crypto import Output as US_OUTPUT

logger = logging.getLogger(__name__)


class BaseCollector(ABC):
    def __init__(self, network) -> None:
        self.data: Optional[Data] = None
        self.network = network

    @abstractmethod
    def is_correct_data_format(self, s) -> bool:
        pass

    @abstractmethod
    def is_complete(self) -> bool:
        pass

    @abstractmethod
    def get_complete_data(self) -> Optional[Data]:
        pass

    @abstractmethod
    def add(self, s: str):
        pass

    def clear(self):
        self.data = None

    @abstractmethod
    def estimated_percent_complete(self):
        pass


class SinglePassCollector(BaseCollector):
    def is_correct_data_format(self, s) -> bool:
        return True

    def is_complete(self) -> bool:
        return bool(self.last)

    def get_complete_data(self) -> Optional[Data]:
        return Data.from_str(self.last, network=self.network)

    def add(self, s: str) -> str:
        self.last = s
        return self.last

    def estimated_percent_complete(self):
        return float(bool(self.data))


class SpecterDIYCollector(BaseCollector):
    def __init__(self, network) -> None:
        super().__init__(network)
        self.clear()
        self.total_parts: int = 0

    def is_correct_data_format(self, s) -> bool:
        return self.extract_specter_diy_qr_part(s) is not None

    def is_complete(self) -> bool:
        return len(self.parts) == self.total_parts

    def get_complete_data(self) -> Optional[Data]:
        if not self.is_complete():
            return None

        total_s = ""
        for i in range(1, self.total_parts + 1):
            total_s += self.parts[i]
        return Data.from_str(total_s, network=self.network)

    def extract_specter_diy_qr_part(self, s) -> Optional[Tuple[int, int, str]]:
        "pMofM something  ->  (M,N,something)"
        pattern = r"^p(\d+)of(\d+)\s(.*)"
        match = re.match(pattern, s)
        if match:
            return int(match.group(1)), int(match.group(2)), match.group(3)
        return None

    def add(self, s: str) -> Optional[str]:
        specter_diy_qr_part = self.extract_specter_diy_qr_part(s)
        if not specter_diy_qr_part:
            return None
        m, n, data = specter_diy_qr_part
        if (self.total_parts is not None) and n != self.total_parts:
            # if n != self.total_parts then, it appears we switched to a different qrcode
            self.clear()

        if self.total_parts is None:
            self.total_parts = n

        self.parts[m] = data
        return data

    def clear(self):
        super().clear()
        self.parts: Dict[int, str] = {}
        self.total_parts = None

    def estimated_percent_complete(self):
        if not self.total_parts:
            return 0
        return min(len(self.parts) / max(self.total_parts, 1), 1)


class URCollector(BaseCollector):
    def __init__(self, network) -> None:
        super().__init__(network)
        self.clear()
        self.last_received_part: Optional[str] = None

    def is_psbt(self, s: str):
        return re.search("^UR:CRYPTO-PSBT/", s, re.IGNORECASE)

    def is_descriptor(self, s: str):
        return re.search("^UR:CRYPTO-OUTPUT/", s, re.IGNORECASE)

    def is_bytes(self, s: str):
        return re.search("^UR:BYTES/", s, re.IGNORECASE)

    def is_correct_data_format(self, s) -> bool:
        if self.is_psbt(s):
            return True
        if self.is_descriptor(s):
            return True
        if self.is_bytes(s):
            return True

        return False

    def is_complete(self) -> bool:
        return self.decoder.is_complete()

    def get_complete_data(self) -> Data:
        if self.decoder.result.type == "crypto-psbt":
            qr_content = UR_PSBT.from_cbor(self.decoder.result.cbor).data
            s = base64.b64encode(qr_content).decode("utf-8")
        if self.decoder.result.type == "crypto-output":
            s = US_OUTPUT.from_cbor(self.decoder.result.cbor).descriptor()
        if self.decoder.result.type == "bytes":
            raw = UR_BYTES.from_cbor(self.decoder.result.cbor).data
            s = raw.hex()

        return Data.from_str(s, network=self.network)

    def add(self, s: str) -> Optional[str]:
        self.decoder.receive_part(s)
        logger.debug(f"{round(self.decoder.estimated_percent_complete()*100)}% complete")

        # if the decoder is stuck for some reason. reset it
        # this part must be done AFTER receive_part(s)
        if self.received_parts > self.decoder.expected_part_count() * 2:
            logger.debug(
                f"{self.received_parts}/{self.decoder.expected_part_count()} parts received, but incomplete. Resetting"
            )
            self.clear()

        if self.last_received_part != s:
            self.received_parts += 1
            self.last_received_part = s
        return s

    def clear(self):
        super().clear()
        self.decoder = URDecoder()
        self.received_parts = 0
        self.last_received_part = None

    def estimated_percent_complete(self):
        return len(self.decoder.received_part_indexes()) / self.decoder.expected_part_count()


class BBQRCollector(BaseCollector):
    def __init__(self, network) -> None:
        super().__init__(network)
        self.clear()

    def clear(self):
        super().clear()
        self.parts: Dict[int, str] = {}
        self.total_parts = None

    def is_correct_data_format(self, s) -> bool:
        if self.get_splitted_data(s):
            return True

        return False

    def get_splitted_data(self, part) -> Optional[Tuple[str, str, int, int, str]]:
        "validation. If invalid, returns None"
        try:
            assert part[0:2] == "B$", "fixed header not found, expected B$"
            encoding = part[2]
            file_type = part[3]
            num_parts = int(part[4:6], 36)

            assert num_parts >= 1, "zero parts?"
            assert encoding in "H2Z", f"bad encoding: {encoding}"
            assert file_type in KNOWN_FILETYPES, f"bad file type: {encoding}"

            idx = int(part[6:8], 36)

            assert idx < num_parts, f"got part {idx} but only expecting {num_parts}"

            raw = part[8:]
            return encoding, file_type, num_parts, idx, raw
        except:
            return None

    def get_available_indices(self) -> Set[int]:

        return set(self.parts.keys())

    def estimated_percent_complete(self):
        num_parts = self.total_parts
        if not num_parts:
            return 0
        return len(self.get_available_indices()) / num_parts

    def is_complete(self) -> bool:
        return self.estimated_percent_complete() >= 1

    def get_complete_data(self) -> Optional[Data]:
        if not self.is_complete():
            return None

        try:
            file_type, raw = bbqr.join_qrs(list(self.parts.values()))
        except:
            self.clear()
            return None

        if FILETYPE_NAMES[file_type] in ["Transaction", "Binary"]:
            return Data.from_binary(raw, network=self.network)
        if FILETYPE_NAMES[file_type] in ["PSBT"]:

            return Data.from_binary(raw, network=self.network)

        if FILETYPE_NAMES[file_type] in ["JSON", "Unicode Text"]:
            return Data.from_str(raw.decode(), network=self.network)

        return None

    def _are_consistent(self, *parts) -> bool:
        if not parts:
            return True
        meta_datas = [self.get_splitted_data(part) for part in parts]
        if not all(meta_datas):
            # Not a single one can be None
            return False

        for i in [0, 1, 2]:
            if len(set([meta_data[i] for meta_data in meta_datas if meta_data])) != 1:
                return False

        return True

    def add(self, s: str) -> Optional[str]:
        # only allow valid parts to be added
        if not self.is_correct_data_format(s):
            return None

        # compare with exising  meta data
        if self.parts:
            existing_part = list(self.parts.values())[0]
            if not self._are_consistent(existing_part, s):
                logger.debug(f"Clearing cache, because {existing_part} if inconsistent with {s}")
                self.clear()

        meta_data = self.get_splitted_data(s)
        if meta_data:
            encoding, file_type, num_parts, idx, raw = meta_data
            self.parts[idx] = s
            self.total_parts = num_parts

        logger.debug(f"{round(self.estimated_percent_complete()*100)}% complete")

        return s


class UnifiedDecoder:
    "Unified class to handle animated and static qr codes"

    def __init__(self, network) -> None:
        self.network = network
        # SinglePassCollector must be the last one
        self.collectors: List[BaseCollector] = [
            BBQRCollector(self.network),
            URCollector(self.network),
            SpecterDIYCollector(self.network),
            SinglePassCollector(self.network),
        ]
        self.last_used_collector: Optional[BaseCollector] = None

    def set_network(self, network: bdk.Network):
        self.network = network
        for collector in self.collectors:
            collector.network = network

    def get_collector(self, s: str) -> Optional[BaseCollector]:
        for collector in self.collectors:
            if collector.is_correct_data_format(s):
                return collector
        return None

    def add(self, s: str):
        self.last_used_collector = self.get_collector(s)
        if not self.last_used_collector:
            raise Exception("Could not get a fitting colletor")
        return self.last_used_collector.add(s)

    def is_complete(self) -> bool:
        if not self.last_used_collector:
            return False
        return self.last_used_collector.is_complete()

    def get_complete_data(self) -> Optional[Data]:
        if not self.last_used_collector:
            return None
        data = self.last_used_collector.get_complete_data()
        self.last_used_collector.clear()
        return data

    def estimated_percent_complete(self):
        if not self.last_used_collector:
            return 0
        return self.last_used_collector.estimated_percent_complete()



#
# crc32.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .constants import MAX_UINT32


def bit_length(n):
    return len(bin(abs(n))) - 2


TABLE = None


def crc32(buf):
    # Lazily instantiate CRC table
    global TABLE
    if TABLE == None:
        TABLE = [None] * (256 * 4)

        for i in range(256):
            c = i
            for j in range(8):
                c = (c >> 1) if (c % 2 == 0) else (0xEDB88320 ^ (c >> 1))

            TABLE[i] = c

    crc = MAX_UINT32 & ~0
    for byte in buf:
        crc = (crc >> 8) ^ TABLE[(crc ^ byte) & 0xFF]

    return MAX_UINT32 & ~crc


def crc32n(buf):
    n = crc32(buf)
    return n.to_bytes(4, "big")



#
# fountain_decoder.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .fountain_utils import choose_fragments, contains, is_strict_subset, set_difference
from .utils import crc32_int, join_bytes, take_first, xor_with


class InvalidPart(Exception):
    pass


class InvalidChecksum(Exception):
    pass


class FountainDecoder:
    class Part:
        def __init__(self, indexes, data):
            self.indexes = frozenset(indexes)
            self.data = data

        @classmethod
        def from_encoder_part(cls, p):
            return cls(choose_fragments(p.seq_num, p.seq_len, p.checksum), p.data[:])

        def indexes(self):
            return self.indexes

        def data(self):
            return self.data

        def is_simple(self):
            return len(self.indexes) == 1

        def index(self):
            # TODO: Not efficient
            return list(self.indexes)[0]

    # FountainDecoder
    def __init__(self):
        self.received_part_indexes = set()
        self.last_part_indexes = None
        self.processed_parts_count = 0
        self.result = None
        self.expected_part_indexes = None
        self.expected_fragment_len = None
        self.expected_message_len = None
        self.expected_checksum = None
        self.simple_parts = {}
        self.mixed_parts = {}
        self.queued_parts = []

    def expected_part_count(self):
        # assume, if there is not expted, that 1 part is all there is
        if self.expected_part_indexes is None:
            return 1
        return len(self.expected_part_indexes)  # TODO: Handle None?

    def is_success(self):
        result = self.result
        return result if not isinstance(result, Exception) else False

    def is_failure(self):
        result = self.result
        return result if isinstance(result, Exception) else False

    def is_complete(self):
        return self.result != None

    def result_message(self):
        return self.result

    def result_error(self):
        return self.result

    def estimated_percent_complete(self):
        if self.is_complete():
            return 1
        if self.expected_part_indexes == None:
            return 0
        estimated_input_parts = self.expected_part_count() * 1.75
        return min(0.99, self.processed_parts_count / estimated_input_parts)

    def receive_part(self, encoder_part):
        # Don't process the part if we're already done
        if self.is_complete():
            return False

        # Don't continue if this part doesn't validate
        if not self.validate_part(encoder_part):
            return False

        # Add this part to the queue
        p = FountainDecoder.Part.from_encoder_part(encoder_part)
        self.last_part_indexes = p.indexes
        self.enqueue(p)

        # Process the queue until we're done or the queue is empty
        while not self.is_complete() and len(self.queued_parts) != 0:
            self.process_queue_item()

        # Keep track of how many parts we've processed
        self.processed_parts_count += 1

        # self.print_part_end()

        return True

    # Join all the fragments of a message together, throwing away any padding
    @staticmethod
    def join_fragments(fragments, message_len):
        message = join_bytes(fragments)
        return take_first(message, message_len)

    def enqueue(self, p):
        self.queued_parts.append(p)

    def process_queue_item(self):
        part = self.queued_parts.pop(0)
        # self.print_part(part)

        if part.is_simple():
            self.process_simple_part(part)
        else:
            self.process_mixed_part(part)
        # self.print_state()

    def reduce_mixed_by(self, p):
        # Reduce all the current mixed parts by the given part
        reduced_parts = []
        for value in self.mixed_parts.values():
            reduced_parts.append(self.reduce_part_by_part(value, p))

        # Collect all the remaining mixed parts
        new_mixed = {}
        for reduced_part in reduced_parts:
            # If this reduced part is now simple
            if reduced_part.is_simple():
                # Add it to the queue
                self.enqueue(reduced_part)
            else:
                # Otherwise, add it to the dict of current mixed parts
                new_mixed[reduced_part.indexes] = reduced_part

        self.mixed_parts = new_mixed

    def reduce_part_by_part(self, a, b):
        # If the fragments mixed into `b` are a strict (proper) subset of those in `a`...
        if is_strict_subset(b.indexes, a.indexes):
            # The new fragments in the revised part are `a` - `b`.
            new_indexes = set_difference(a.indexes, b.indexes)
            # The new data in the revised part are `a` XOR `b`
            new_data = xor_with(bytearray(a.data), b.data)
            return self.Part(new_indexes, new_data)
        else:
            # `a` is not reducable by `b`, so return a
            return a

    def process_simple_part(self, p):
        # Don't process duplicate parts
        fragment_index = p.index()
        if contains(self.received_part_indexes, fragment_index):
            return

        # Record this part
        self.simple_parts[p.indexes] = p
        self.received_part_indexes.add(fragment_index)

        # If we've received all the parts
        if self.received_part_indexes == self.expected_part_indexes:
            # Reassemble the message from its fragments
            sorted_parts = []
            for value in self.simple_parts.values():
                sorted_parts.append(value)

            sorted_parts.sort(key=lambda a: a.index())

            fragments = []
            for part in sorted_parts:
                fragments.append(part.data)

            message = self.join_fragments(fragments, self.expected_message_len)

            # Verify the message checksum and note success or failure
            checksum = crc32_int(message)
            if checksum == self.expected_checksum:
                self.result = bytes(message)
            else:
                self.result = InvalidChecksum()

        else:
            # Reduce all the mixed parts by this part
            self.reduce_mixed_by(p)

    def process_mixed_part(self, p):
        # Don't process duplicate parts
        for r in self.mixed_parts.values():
            if r == p.indexes:
                return

        # Reduce this part by all the others
        p2 = p  # TODO: Does this need to make a copy of p?
        for r in self.simple_parts.values():
            p2 = self.reduce_part_by_part(p2, r)

        for r in self.mixed_parts.values():
            p2 = self.reduce_part_by_part(p2, r)

        # If the part is now simple
        if p2.is_simple():
            # Add it to the queue
            self.enqueue(p2)
        else:
            # Reduce all the mixed parts by this one
            self.reduce_mixed_by(p2)
            # Record this new mixed part
            self.mixed_parts[p2.indexes] = p2

    def validate_part(self, p):
        # If this is the first part we've seen
        if self.expected_part_indexes == None:
            # Record the things that all the other parts we see will have to match to be valid.
            self.expected_part_indexes = set()
            for i in range(p.seq_len):
                self.expected_part_indexes.add(i)

            self.expected_message_len = p.message_len
            self.expected_checksum = p.checksum
            self.expected_fragment_len = len(p.data)
        else:
            # If this part's values don't match the first part's values, throw away the part
            if self.expected_part_count() != p.seq_len:
                return False
            if self.expected_message_len != p.message_len:
                return False
            if self.expected_checksum != p.checksum:
                return False
            if self.expected_fragment_len != len(p.data):
                return False

        # This part should be processed
        return True

    # debugging
    def indexes_to_string(self, indexes):
        i = list(indexes)
        i.sort()
        s = [str(j) for j in i]
        return "[{}]".format(", ".join(s))

    def result_description(self):
        if self.result == None:
            return "None"

        if self.is_success():
            return "{} bytes".format(len(self.result))
        elif self.is_failure():
            return "Exception: {}".format(self.result)
        else:
            assert False

    def print_part(self, p):
        print("part indexes: {}".format(self.indexes_to_string(p.indexes)))

    def print_part_end(self):
        expected = self.expected_part_count() if self.expected_part_indexes != None else "None"
        percent = int(round(self.estimated_percent_complete() * 100))
        print(
            "processed: {}, expected: {}, received: {}, percent: {}%".format(
                self.processed_parts_count,
                expected,
                len(self.received_part_indexes),
                percent,
            )
        )

    def print_state(self):
        parts = self.expected_part_count() if self.expected_part_indexes != None else "None"
        received = self.indexes_to_string(self.received_part_indexes)
        mixed = []
        for indexes, p in self.mixed_parts.items():
            mixed.append(self.indexes_to_string(indexes))

        mixed_s = "[{}]".format(", ".join(mixed))
        queued = len(self.queued_parts)
        res = self.result_description()
        print(
            "parts: {}, received: {}, mixed: {}, queued: {}, result: {}".format(
                parts, received, mixed_s, queued, res
            )
        )



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun
# Copyright (c) 2015 Sokolov Yura
# Copyright (c) 2013 Fritz Grimpen

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# coding: utf-8

import struct

_str_type = type("")
_bytes_type = type(b"")

from .data import Mapping, Tagging, Undefined


class EncoderError(Exception):
    pass


class Encoder(object):
    def __init__(self, output):
        self.output = output

    def encode(self, val):
        if isinstance(val, _bytes_type):
            self.encode_bytestring(val)
        elif isinstance(val, _str_type):
            self.encode_textstring(val)
        elif isinstance(val, float):
            self.encode_float(val)
        elif isinstance(val, bool):
            self.encode_boolean(val)
        elif isinstance(val, int):
            self.encode_integer(val)
        elif isinstance(val, list):
            self.encode_list(val)
        elif isinstance(val, dict):
            self.encode_dict(val)
        elif isinstance(val, Tagging):
            self.encode_tagging(val)
        elif val is Undefined:
            self.encode_undefined()
        elif val is None:
            self.encode_null()
        elif isinstance(val, Mapping):
            val = val.map
            self.encode(val)
        else:
            raise EncoderError("val of type {} is not serializable".format(type(val)))

    def encode_list(self, list):
        self._write(_encode_ibyte(4, len(list)))
        for elem in list:
            self.encode(elem)

    def encode_dict(self, dict):
        self._write(_encode_ibyte(5, len(dict)))
        for key, value in dict.items():
            self.encode(key)
            self.encode(value)

    def encode_bytestring(self, bytestring):
        self._write(_encode_ibyte(2, len(bytestring)))
        self._write(bytestring)

    def encode_textstring(self, textstring):
        string_ = textstring.encode("utf-8")
        self._write(_encode_ibyte(3, len(string_)))
        self._write(string_)

    def encode_float(self, float):
        self._write(b"\xfb")
        self._write(struct.pack(">d", float))

    def encode_integer(self, integer):
        if integer < 0:
            integer = -integer - 1
            try:
                self._write(_encode_ibyte(1, integer))
            except TypeError:
                raise EncoderError("Encoding integers lower than -18446744073709551616 is not supported")
        else:
            try:
                self._write(_encode_ibyte(0, integer))
            except TypeError:
                raise EncoderError("Encoding integers larger than 18446744073709551615 is not supported")

    def encode_tagging(self, tagging):
        try:
            self._write(_encode_ibyte(6, tagging.tag))
        except TypeError:
            raise EncoderError("Encoding tag larger than 18446744073709551615 is not supported")
        self.encode(tagging.obj)

    def encode_boolean(self, boolean):
        if boolean is True:
            self._write(_encode_ibyte(7, 21))
        elif boolean is False:
            self._write(_encode_ibyte(7, 20))

    def encode_null(self):
        self._write(_encode_ibyte(7, 22))

    def encode_undefined(self):
        self._write(b"\xf7")

    def _write(self, val):
        self.output.write(val)


def _encode_ibyte(major, length):
    if length < 24:
        return struct.pack(">B", (major << 5) | length)
    elif length < 256:
        return struct.pack(">BB", (major << 5) | 24, length)
    elif length < 65536:
        return struct.pack(">BH", (major << 5) | 25, length)
    elif length < 4294967296:
        return struct.pack(">BI", (major << 5) | 26, length)
    elif length < 18446744073709551616:
        return struct.pack(">BQ", (major << 5) | 27, length)
    else:
        return None


__all__ = ["Encoder", "EncoderError"]



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ..urtypes import RegistryItem, RegistryType

BYTES = RegistryType("bytes", None)


class Bytes(RegistryItem):
    def __init__(self, data):
        super().__init__()
        self.data = data

    def __eq__(self, o):
        return self.data == o.data

    @classmethod
    def registry_type(cls):
        return BYTES

    def to_data_item(self):
        return self.data

    @classmethod
    def from_data_item(cls, item):
        return cls(cls.mapping(item))



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import io
import logging
import sys
from typing import List, Optional, Tuple

from PIL import Image
from PyQt6.QtCore import QByteArray, QEvent, QRectF, QSize, Qt, QTimer
from PyQt6.QtGui import (
    QCloseEvent,
    QImage,
    QKeyEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QPixmap,
)
from PyQt6.QtSvg import QSvgRenderer
from PyQt6.QtWidgets import QApplication, QDialog, QWidget

from bitcoin_qr_tools.data import Data
from bitcoin_qr_tools.qr_generator import QRGenerator

logger = logging.getLogger(__name__)


def pil_image_to_qimage(im: Image):
    im = im.convert("RGBA")
    data = im.tobytes("raw", "RGBA")
    qim = QImage(data, im.size[0], im.size[1], QImage.Format.Format_RGBA8888)

    return qim.copy()  # Making a copy to let data persist after function returns


class ImageWidget(QWidget):
    def __init__(self, pil_image: Image = None, parent=None, size_hint: Tuple[int, int] = None):
        super().__init__(parent)
        self.pil_image = pil_image
        self.size_hint = size_hint
        self.qt_image = pil_image_to_qimage(pil_image) if pil_image else QImage()
        self.scaled_image = self.qt_image

    def paintEvent(self, event: QPaintEvent) -> None:
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, False)

        if not self.qt_image.isNull():
            widget_width, widget_height = self.width(), self.height()

            # choose minimum of image and self sizes
            width = min(self.qt_image.size().width(), widget_width)
            height = min(self.qt_image.size().height(), widget_height)

            # Scale the image to fit within the widget while maintaining aspect ratio
            self.scaled_image = self.qt_image.scaled(
                width, height, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation
            )

            # Calculate position to center the image
            x = (widget_width - self.scaled_image.width()) // 2
            y = (widget_height - self.scaled_image.height()) // 2

            # Draw the image centered
            painter.drawImage(x, y, self.scaled_image)

    def set_image(self, pil_image: Image):
        self.pil_image = pil_image
        self.qt_image = pil_image_to_qimage(pil_image)
        self.update()  # Trigger a repaint

    def load_from_file(self, filepath: str):
        self.set_image(Image.open(filepath))

    def sizeHint(self) -> QSize:
        if not self.qt_image.isNull():
            if not self.size_hint:
                return self.qt_image.size()
            else:
                s = QSize()
                s.setWidth(self.size_hint[0])
                s.setHeight(self.size_hint[1])
                return s
        return super().sizeHint()


class EnlargableImageWidget(ImageWidget):
    def __init__(self, pil_image: Image = None, parent=None, size_hint: Tuple[int, int] = None):
        super().__init__(pil_image, parent, size_hint=size_hint)
        self.enlarged_image: Optional[EnlargedImage] = None
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def enlarge_image(self):
        if not self.enlarged_image:
            self.enlarged_image = EnlargedImage(self.pil_image)

        if self.enlarged_image.isVisible():
            self.enlarged_image.close()
        else:
            self.enlarged_image.show()

    def mousePressEvent(self, event: QMouseEvent):
        self.enlarge_image()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.enlarged_image:
            self.enlarged_image.close()
        super().closeEvent(event)


class EnlargedImage(ImageWidget):
    def __init__(self, pil_image: Image, parent=None, screen_fraction=0.4):
        super().__init__(pil_image, parent)
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.installEventFilter(self)  # Install the event filter for this widget

        # Get screen resolution
        screen = QApplication.screens()[0].size()

        # Calculate the new size maintaining the aspect ratio
        image_aspect_ratio = self.qt_image.width() / self.qt_image.height()
        new_width = min(screen.width() * screen_fraction, self.qt_image.width())
        new_height = new_width / image_aspect_ratio

        # Ensure the height does not exceed 50% of the screen height
        if new_height > screen.height() * screen_fraction:
            new_height = screen.height() * screen_fraction
            new_width = new_height * image_aspect_ratio

        # Calculate position to center the window
        x = round((screen.width() - new_width) / 2)
        y = round((screen.height() - new_height) / 2)

        self.setGeometry(x, y, round(new_width), round(new_height))

    def eventFilter(self, source: QWidget, event: QEvent) -> bool:
        # Check for the FocusOut event
        if event.type() in [QEvent.Type.FocusOut, QEvent.Type.WindowDeactivate]:
            # Close the widget if it loses focus
            if source is self:
                self.close()
        return super().eventFilter(source, event)

    def mousePressEvent(self, event: QMouseEvent):
        self.close()

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key.Key_Escape:
            self.close()


class QRCodeWidget(EnlargableImageWidget):
    def __init__(self, parent=None, clickable=True):
        super().__init__(parent=parent)
        # QR code specific initializations, if any

    def set_data(self, data: str):
        # Implement QR code generation and setting image
        self.set_image(QRGenerator.create_qr_PILimage(data))


#######################################
# svg widgets


class QRCodeWidgetSVG(QWidget):
    def __init__(self, always_animate=False, clickable=True, parent=None):
        super().__init__(parent)
        self.svg_renderers: List[QSvgRenderer] = []
        self.current_index = 0
        self.enlarged_image = None
        self.clickable = clickable
        self.always_animate = always_animate
        self.is_hovered = False
        self.default_size = 200

        if clickable:
            self.setCursor(Qt.CursorShape.PointingHandCursor)

        self.timer = QTimer()
        self.timer.timeout.connect(self.next_svg)

    def sizeHint(self) -> QSize:
        s = QSize()
        s.setWidth(self.default_size)
        s.setHeight(self.default_size)
        return s

    def set_data_list(self, data_list: List[str]):
        self.svg_renderers = [
            QSvgRenderer(QByteArray(QRGenerator.create_qr_svg(data).encode("utf-8"))) for data in data_list
        ]
        self.current_index = 0
        self.manage_animation()

    def set_always_animate(self, always_animate: bool):
        self.always_animate = always_animate
        self.manage_animation()

    def set_images(self, image_list: List[str]):
        self.svg_renderers = [QSvgRenderer(QByteArray(image.encode("utf-8"))) for image in image_list]
        self.current_index = 0
        self.manage_animation()
        self.update()

    def manage_animation(self):
        should_animate = len(self.svg_renderers) > 1 and (
            self.always_animate
            or self.is_hovered
            or (self.enlarged_image and self.enlarged_image.isVisible())
        )
        if should_animate:
            self.timer.start(1000)  # Change SVG every 1 second
        else:
            self.timer.stop()

    def next_svg(self):
        if not self.svg_renderers:
            return

        self.current_index = (self.current_index + 1) % len(self.svg_renderers)
        self.update()
        if self.enlarged_image and self.enlarged_image.isVisible():
            self.enlarged_image.update_image(self.svg_renderers[self.current_index])
        else:
            self.manage_animation()

    def paintEvent(self, event: QPaintEvent) -> None:
        if not self.svg_renderers:
            return

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        widget_width, widget_height = self.width(), self.height()
        side = min(widget_width, widget_height)
        x = (widget_width - side) // 2
        y = (widget_height - side) // 2

        self.svg_renderers[self.current_index].render(painter, QRectF(x, y, side, side))

    def enterEvent(self, event: QEvent) -> None:
        self.is_hovered = True
        self.manage_animation()

    def leaveEvent(self, event: QEvent) -> None:
        self.is_hovered = False
        self.manage_animation()

    def enlarge_image(self):
        if not self.svg_renderers:
            return

        if not self.enlarged_image:
            self.is_hovered = False
            self.enlarged_image = EnlargedSVG(self.svg_renderers[self.current_index])

        self.enlarged_image.exec()
        self.enlarged_image.update_image(self.svg_renderers[self.current_index])
        self.manage_animation()

    def mousePressEvent(self, event: QMouseEvent):
        if self.clickable:
            self.enlarge_image()

    def save_file(self, base_filename: str, format="PNG", antialias=False):
        """Save all QR codes to files. If format is 'GIF', combines them into
        an animated GIF.

        :param base_filename: Base path and filename without extension.
        :param format: The format in which to save the image (e.g.,
            'PNG', 'GIF').
        :param antialias: Boolean to indicate if anti-aliasing should be
            used.
        """
        if format.upper() == "GIF":
            images = []
            for renderer in self.svg_renderers:
                if not renderer.isValid():
                    continue
                images.append(self.renderer_to_pil(renderer, antialias))
            images[0].save(
                f"{base_filename}.gif",
                save_all=True,
                append_images=images[1:],
                loop=0,
                duration=1000,
            )
        else:
            for i, renderer in enumerate(self.svg_renderers):
                if not renderer.isValid():
                    continue
                image = self.renderer_to_pil(renderer, antialias)
                image.save(f"{base_filename}_{i}.{format.lower()}")

    def renderer_to_pil(self, renderer: QSvgRenderer, antialias: bool):
        """Convert a QR code renderer to a PIL Image.

        :param renderer: The QR code renderer.
        :param antialias: Boolean to indicate if anti-aliasing should be
            used.
        :return: PIL Image object.
        """
        size = self.size()
        pixmap = QPixmap(size)
        pixmap.fill(Qt.GlobalColor.white)
        painter = QPainter(pixmap)

        if antialias:
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        renderer.render(painter, QRectF(0, 0, size.width(), size.height()))
        painter.end()

        # Convert QPixmap to QImage
        qimage = pixmap.toImage()

        # Convert QImage to PIL Image
        buffer = io.BytesIO()
        qimage.save(buffer, "PNG")
        buffer.seek(0)
        return Image.open(buffer)

    def as_pil_images(self):
        """Convert all the QR codes to PIL Images.

        :return: List of PIL Image objects.
        """
        return [
            self.renderer_to_pil(renderer, antialias=True)
            for renderer in self.svg_renderers
            if renderer.isValid()
        ]

    def closeEvent(self, event: QCloseEvent) -> None:
        self.timer.stop()
        self.enlarged_image.close()
        super().closeEvent(event)


class EnlargedSVG(QDialog):
    def __init__(self, svg_renderer: QSvgRenderer, parent=None, screen_fraction=0.5):
        super().__init__(parent)
        self.svg_renderer = svg_renderer
        self.setWindowTitle("QR Code")
        # self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint)
        screen = QApplication.screens()[0].size()
        width = height = round(min(screen.width(), screen.height()) * screen_fraction)
        self.setGeometry(
            round((screen.width() - width) / 2),
            round((screen.height() - height) / 2),
            width,
            height,
        )

    def update_image(self, new_renderer: QSvgRenderer):
        self.svg_renderer = new_renderer
        self.update()

    def paintEvent(self, event: QPaintEvent) -> None:
        if not self.svg_renderer:
            return

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        widget_width, widget_height = self.width(), self.height()
        side = min(widget_width, widget_height)
        x = (widget_width - side) // 2
        y = (widget_height - side) // 2

        self.svg_renderer.render(painter, QRectF(x, y, side, side))

    def mousePressEvent(self, event: QMouseEvent):
        super().mousePressEvent(event)
        self.close()

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key.Key_Escape:
            self.close()


if __name__ == "__main__":
    import bdkpython as bdk

    app = QApplication(sys.argv)
    widget = QRCodeWidgetSVG()

    psbt = "cHNidP8BAHEBAAAAAXgQzjk+DTWQTPUtRMbYiheC0jfbipvw+jQ5lidmyABjAAAAAAD9////AgDh9QUAAAAAFgAUbBuOQOlcnz8vpruh2Kb3CFr4vlhkEQ2PAAAAABYAFN1n2hvBWYzshD42xwQzy9XYoji3BAEAAAABAKoCAAAAAAEBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD/////BQKYAAEB/////wIA+QKVAAAAABYAFLlHwN6VXNLM381bMxmNJlaDTQzVAAAAAAAAAAAmaiSqIant4vYcP3HR3v0/qZnfo2lTdVxpBol5mWK0i+vYNpdOjPkBIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEBHwD5ApUAAAAAFgAUuUfA3pVc0szfzVszGY0mVoNNDNUiBgISCnRxeOxzC0MgK01AmiIRLrgS1AyIqKeBkdwL+nt/6RikLG3TVAAAgAEAAIAAAACAAAAAAAAAAAAAACICAlQcwExiTUk9f7olLkwPlQpiregRHc9jXXFJBlMoucgNGKQsbdNUAACAAQAAgAAAAIAAAAAAAQAAAAA="
    data: Data = Data.from_str(psbt, network=bdk.Network.REGTEST)
    fragments = data.generate_fragments_for_qr(qr_type="bbqr")

    widget.set_data_list(fragments)
    widget.show()

    sys.exit(app.exec())



#
# (c) Copyright 2023 by Coinkite Inc. This file is in the public domain.
#
# - helpers and basics
#
import zlib
from base64 import b32decode, b32encode


def version_to_chars(v):
    # return number of **chars** that fit into indicated version QR
    # - assumes L for ECC
    # - assumes alnum encoding
    import pyqrcode

    assert 1 <= v <= 40
    ecc = "L"
    encoding = 2  # alnum

    return pyqrcode.tables.data_capacity[v][ecc][encoding]


def int2base36(n):
    # convert an integer to two digits of base 36 string. 00 thu ZZ
    # converse is just int(s, base=36)
    assert 0 <= n <= 1295

    tostr = lambda x: chr(48 + x) if x < 10 else chr(65 + x - 10)

    a, b = divmod(n, 36)

    return tostr(a) + tostr(b)


def encode_data(raw, encoding=None):
    # return new encoding (if we upgraded) and the
    # characters after encoding (a string)
    # - default is Zlib or if compression doesn't help, base32
    # - returned data can be split, but must be done modX where X provided

    if encoding == "H":
        # Hex mode is easy.
        return encoding, raw.hex().upper(), 2

    if not encoding or encoding == "Z":
        # Trial compression, but skip if it embiggens the data
        z = zlib.compressobj(wbits=-10)
        cmp = z.compress(raw)
        cmp += z.flush()
        if len(cmp) >= len(raw):
            encoding = "2"
        else:
            encoding = "Z"
            raw = cmp

    # Default: base32 encoding, no padding bytes
    data = b32encode(raw).decode("ascii").rstrip("=")

    return encoding, data, 8


def decode_data(parts, encoding):
    # give back the bytes after decoding
    # - already in order
    # - keeps the parts separate here to validate correct split from encoder
    if encoding == "H":
        return b"".join(bytes.fromhex(p) for p in parts)

    # base32 decode, but insert padding for API
    rv = b""
    for p in parts:
        padding = (8 - (len(p) % 8)) % 8
        rv += b32decode(p + (padding * "="))

    if encoding == "Z":
        # decompress
        z = zlib.decompressobj(wbits=-10)
        rv = z.decompress(rv)
        rv += z.flush()

    return rv


# EOF



#
# (c) Copyright 2023 by Coinkite Inc. This file is put in the public domain.
#
# - just two funcitons you should need
#

from .join import join_qrs
from .split import split_qrs
from .version import __version__



#
# (c) Copyright 2023 by Coinkite Inc. This file is in the public domain.
#
#
# Constants and fixed values
#

# Standard defines a fixed-length header
HEADER_LEN = 8

# Human names
FILETYPE_NAMES = dict(
    P="PSBT", T="Transaction", J="JSON", C="CBOR", U="Unicode Text", X="Executable", B="Binary"
)

# Codes for PSBT vs. TXN and so on
KNOWN_FILETYPES = set(FILETYPE_NAMES.keys())

# EOF



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import binascii
import hashlib

from ...urtypes import RegistryItem, RegistryType
from ...urtypes.cbor import DataItem
from .coin_info import CoinInfo
from .keypath import Keypath

CRYPTO_HDKEY = RegistryType("crypto-hdkey", 303)


class HDKey(RegistryItem):
    def __init__(self, props):
        super().__init__()
        self.master = None
        self.key = None
        self.chain_code = None
        self.private_key = None
        self.use_info = None
        self.origin = None
        self.children = None
        self.parent_fingerprint = None
        self.name = None
        self.note = None
        if "master" in props and props["master"]:
            self.setup_master_key(props)
        else:
            self.setup_derive_key(props)

    def __eq__(self, o):
        return (
            self.master == o.master
            and self.key == o.key
            and self.chain_code == o.chain_code
            and self.private_key == o.private_key
            and self.use_info == o.use_info
            and self.origin == o.origin
            and self.children == o.children
            and self.parent_fingerprint == o.parent_fingerprint
            and self.name == o.name
            and self.note == o.note
        )

    @classmethod
    def registry_type(cls):
        return CRYPTO_HDKEY

    def setup_master_key(self, props):
        self.master = True
        self.key = props["key"] if "key" in props else None
        self.chain_code = props["chain_code"] if "chain_code" in props else None

    def setup_derive_key(self, props):
        self.master = False
        self.key = props["key"] if "key" in props else None
        self.chain_code = props["chain_code"] if "chain_code" in props else None
        self.private_key = props["private_key"] if "private_key" in props else None
        self.use_info = props["use_info"] if "use_info" in props else None
        self.origin = props["origin"] if "origin" in props else None
        self.children = props["children"] if "children" in props else None
        self.parent_fingerprint = props["parent_fingerprint"] if "parent_fingerprint" in props else None
        self.name = props["name"] if "name" in props else None
        self.note = props["note"] if "note" in props else None

    def bip32_key(self, include_derivation_path=False):
        parent_fingerprint = (0).to_bytes(4, "big")
        source_is_parent = False
        chain_code = self.chain_code if self.chain_code is not None else (0).to_bytes(32, "big")
        key = self.key
        if len(key) == 32:
            key = 0x00 + key
        depth = 0
        index = 0
        if self.master:
            version = binascii.unhexlify(
                "0488ADE4" if not self.use_info or self.use_info.network == 0 else "04358394"
            )
        else:
            if self.private_key:
                version = binascii.unhexlify(
                    "0488ADE4" if not self.use_info or self.use_info.network == 0 else "04358394"
                )
            else:
                version = binascii.unhexlify(
                    "0488B21E" if not self.use_info or self.use_info.network == 0 else "043587CF"
                )
            if self.parent_fingerprint is not None:
                parent_fingerprint = self.parent_fingerprint
            depth = self.origin.depth if self.origin.depth is not None else len(self.origin.components)
            paths = self.origin.components
            if len(paths) > 0:
                last_path = paths[len(paths) - 1]
                index = last_path.index
                if last_path.hardened:
                    index += 0x80000000
                if (
                    self.parent_fingerprint is None
                    and self.origin.source_fingerprint is not None
                    and len(paths) == 1
                ):
                    parent_fingerprint = self.origin.source_fingerprint
                    source_is_parent = True
        depth = depth.to_bytes(1, "big")
        index = index.to_bytes(4, "big")
        key = encode_check(version + depth + parent_fingerprint + index + chain_code + key)
        if include_derivation_path:
            derivation = ""
            if self.origin and self.origin.path() and self.origin.source_fingerprint and not source_is_parent:
                derivation = "[%s/%s]" % (
                    binascii.hexlify(self.origin.source_fingerprint).decode("utf-8"),
                    self.origin.path(),
                )

            child_derivation = ""
            if self.children and self.children.path():
                child_derivation = "/" + self.children.path()

            return "%s%s%s" % (derivation, key, child_derivation)
        return key

    def descriptor_key(self):
        return self.bip32_key(True)

    def to_data_item(self):
        map = {}
        if self.master:
            map[1] = True
            map[3] = self.key
            map[4] = self.chain_code
        else:
            if self.private_key is not None:
                map[2] = self.private_key
            map[3] = self.key
            if self.chain_code is not None:
                map[4] = self.chain_code
            if self.use_info is not None:
                map[5] = DataItem(self.use_info.registry_type().tag, self.use_info.to_data_item())
            if self.origin is not None:
                map[6] = DataItem(self.origin.registry_type().tag, self.origin.to_data_item())
            if self.children is not None:
                map[7] = DataItem(self.children.registry_type().tag, self.children.to_data_item())
            if self.parent_fingerprint is not None:
                map[8] = int.from_bytes(self.parent_fingerprint, "big")
            if self.name is not None:
                map[9] = self.name
            if self.note is not None:
                map[10] = self.note
        return map

    @classmethod
    def from_data_item(cls, item):
        map = cls.mapping(item)
        master = 1 in map and map[1]
        private_key = map[2] if 2 in map else None
        key = map[3] if 3 in map else None
        chain_code = map[4] if 4 in map else None
        use_info = CoinInfo.from_data_item(map[5]) if 5 in map else None
        origin = Keypath.from_data_item(map[6]) if 6 in map else None
        children = Keypath.from_data_item(map[7]) if 7 in map else None
        parent_fingerprint = map[8].to_bytes(4, "big") if 8 in map else None
        name = map[9] if 9 in map else None
        note = map[10] if 10 in map else None
        return cls(
            {
                "master": master,
                "private_key": private_key,
                "key": key,
                "chain_code": chain_code,
                "use_info": use_info,
                "origin": origin,
                "children": children,
                "parent_fingerprint": parent_fingerprint,
                "name": name,
                "note": note,
            }
        )


B58_DIGITS = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def double_sha256(msg):
    """sha256(sha256(msg)) -> bytes"""
    return hashlib.sha256(hashlib.sha256(msg).digest()).digest()


def encode(b):
    """Encode bytes to a base58-encoded string"""

    # Convert big-endian bytes to integer
    n = int("0x0" + binascii.hexlify(b).decode("utf8"), 16)

    # Divide that integer into bas58
    res = []
    while n > 0:
        n, r = divmod(n, 58)
        res.append(B58_DIGITS[r])
    res = "".join(res[::-1])

    pad = 0
    for c in b:
        if c == 0:
            pad += 1
        else:
            break
    return B58_DIGITS[0] * pad + res


def encode_check(b):
    """Encode bytes to a base58-encoded string with a checksum"""
    return encode(b + double_sha256(b)[0:4])



import logging
from typing import List, Optional

import bdkpython as bdk
from hwilib.descriptor import Descriptor, PubkeyProvider, parse_descriptor

logger = logging.getLogger(__name__)


def get_all_pubkey_providers(hwi_descriptor: Descriptor) -> List[PubkeyProvider]:
    pubkey_providers = hwi_descriptor.pubkeys.copy()
    for subdescriptor in hwi_descriptor.subdescriptors:
        pubkey_providers += get_all_pubkey_providers(subdescriptor).copy()
    return pubkey_providers


def get_adapted_hwi_descriptor(descriptor_str: str, new_derivation_path: str) -> Descriptor:
    hwi_descriptor = parse_descriptor(descriptor_str)
    pubkey_providers = get_all_pubkey_providers(hwi_descriptor)
    for pubkey_provider in pubkey_providers:
        pubkey_provider.deriv_path = new_derivation_path
    return hwi_descriptor


class MultipathDescriptor:
    """
    Will create main+change BDK single and multisig descriptors, no matter if '/<0;1>/*' or '/0/*' or '/1/*' is specified
    It also uses hwi to handle edge cases that bdk doesnt support yet.

    This is only necessary until https://github.com/bitcoindevkit/bdk/issues/1021  is done.
    """

    def __init__(self, bdk_descriptor: bdk.Descriptor, change_descriptor: bdk.Descriptor) -> None:
        self.bdk_descriptors = [bdk_descriptor, change_descriptor]

        for bdk_descriptor in self.bdk_descriptors:
            # check that the self.bdk_descriptors each have equal derivation_paths
            derivation_path = self.get_equal_derivation_path(bdk_descriptor.as_string())
            assert (
                derivation_path
            ), f"Derivation paths in {bdk_descriptor.as_string()} are not all equal. MultipathDescriptor does not  support this."

    @classmethod
    def get_equal_derivation_path(cls, descriptor_str: str) -> Optional[str]:
        "Returns the derivation_path is all derivation_paths are equal. Otherwise None"

        hwi_descriptor = parse_descriptor(descriptor_str)
        pubkey_providers = get_all_pubkey_providers(hwi_descriptor=hwi_descriptor)

        # check that all derivation paths are equal
        derivation_paths = [pubkey_provider.deriv_path for pubkey_provider in pubkey_providers]
        all_equal = all(x == derivation_paths[0] for x in derivation_paths)

        if all_equal:
            return pubkey_providers[0].deriv_path
        else:
            return None

    @classmethod
    def from_descriptor_str(cls, descriptor_str: str, network: bdk.Network) -> "MultipathDescriptor":
        derivation_path = cls.get_equal_derivation_path(descriptor_str)

        assert (
            derivation_path
        ), f"Derivation paths are not all equal, and from this no MultiPathDescriptor can be created."

        # sparrow qr code misses the change derivation path completely
        assert derivation_path in [
            "",
            "/0/*",
            "/1/*",
            "/<0;1>/*",
        ], f"Unknown derivation path {derivation_path}, and from this no MultiPathDescriptor can be created."

        receive_descriptor = get_adapted_hwi_descriptor(descriptor_str, new_derivation_path="/0/*")
        change_descriptor = get_adapted_hwi_descriptor(descriptor_str, new_derivation_path="/1/*")

        return MultipathDescriptor(
            bdk.Descriptor(receive_descriptor.to_string(), network=network),
            bdk.Descriptor(change_descriptor.to_string(), network=network),
        )

    def as_string(self) -> str:
        return self._as_string(only_public=True)

    def as_string_private(self) -> str:
        return self._as_string(only_public=False)

    def _as_string(self, only_public=False) -> str:
        # TODO: Once https://github.com/bitcoindevkit/bdk/issues/1021" solved replace hwi with bdk
        assert len(self.bdk_descriptors) == 2

        receive_descriptor_str = (
            self.bdk_descriptors[0].as_string()
            if only_public
            else self.bdk_descriptors[0].as_string_private()
        )

        receive_descriptor = get_adapted_hwi_descriptor(
            receive_descriptor_str, new_derivation_path="/<0;1>/*"
        )
        return receive_descriptor.to_string(hardened_char="'")



#
# (c) Copyright 2023 by Coinkite Inc. This file is in the public domain.
#
# - uses pyqrcode for deep QR knowledge
# - text prefix on each QR:
#
#       B$                  fixed header for this protocol (2 chars)
#       2                   one char encode: Z=zlib, 2=Base32, H=Hex
#       P                   one char file type: P=PSBT, T=TXN, etc
#       05                  2-digits of HEX: total number of QR codes
#       00                  2-digits of HEX: which QR code this is in the sequence
#
from math import ceil

from .consts import HEADER_LEN, KNOWN_FILETYPES
from .utils import encode_data, int2base36, version_to_chars


def num_qr_needed(ver, ll, split_mod):
    # Determine number of QR's at indicated version would be
    # needed to hold ll characters. when 2 or more QR, consider
    # the exact split point cannot be between encoded symbols
    # - ok to return huge numbers for unlikely cases
    cap = version_to_chars(ver) - HEADER_LEN
    cap2 = cap - (cap % split_mod)

    need = ceil(ll / cap2)

    if need == 1:
        # no alignment concerns
        assert ll <= cap
        return 1, ll

    # going to be 2 or more, gotta be precise
    actual = ((need - 1) * cap2) + cap

    return (need if actual >= ll else (need + 1)), cap2


def find_best_version(ll, split_mod, min_split=1, max_split=1295, min_version=5, max_version=40):
    # Find ideal QR version and provide # of QR and splits needed.
    # - assumes you want to pack the QR, so forcing min_split means you need to have the data
    #   at least the data to fill that # of QR at min_version
    #
    # ll = length of encoded data to be transmitted (no headers)
    # split_mod = required size of non-runt parts so that can be decoded w/o spliting symbols

    min_version = min(min_version, max_version)  # in case they spec a very low max

    assert 1 <= min_version <= max_version <= 40, "min/max version out of range"
    assert 1 <= min_split <= max_split <= 1295, "num splits out of range"

    options = []
    for ver in range(min_version, max_version + 1):
        count, pe = num_qr_needed(ver, ll, split_mod)
        if not (min_split <= count <= max_split):
            continue
        options.append((ver, count, pe))

    # pick smallest number of QR, lowest version
    options.sort(key=lambda v: (v[1], v[0]))

    if not options:
        raise ValueError("Cannot make it fit")

    return options[0]


def split_qrs(raw, type_code, encoding=None, **kws):
    # Take some bytes and yield a series of text values that
    # can be sent as QR code.
    # - returns text
    # - assumes and requires alnum, L error level
    # - see find_best_version() for additional kw args

    assert type_code in KNOWN_FILETYPES, f"invalid type_code: {type_code}"
    if encoding:
        assert encoding in "H2Z", f"invalid encoding: {encoding}"
    if not isinstance(raw, bytes):
        assert isinstance(raw, str), "need binary or text"
        raw = raw.encode("utf-8")

    # perhaps compress data
    encoding, encoded, split_mod = encode_data(raw, encoding)

    ll = len(encoded)

    ver, num_qr, per_each = find_best_version(ll, split_mod, **kws)

    assert per_each * num_qr >= ll

    return ver, [
        f"B${encoding}{type_code}" + int2base36(num_qr) + int2base36(n) + encoded[off : off + per_each]
        for (n, off) in enumerate(range(0, ll, per_each))
    ]


# EOF



#
# (c) Copyright 2023 by Coinkite Inc. This file is in the public domain.
#
# Print a table used in the spec. Use "bbqr table" to view.
#
from math import floor

import pyqrcode

from .consts import HEADER_LEN
from .utils import version_to_chars


def dump_table():
    ver_size = pyqrcode.tables.version_size

    hdr = "Vers | Pixels  | Chars |  Hex |  Base32 | 2xBase32 | 5xBase32 | 10xBase32"
    print(hdr)
    print("|".join("-" * len(i) for i in hdr.split("|")))

    for v in range(1, 41):
        chars = version_to_chars(v)
        if chars < 1500 and v not in {1, 11, 14}:
            continue

        sz = ver_size[v]
        cap = chars - HEADER_LEN
        bys = floor(cap / 2)  # HEX encoding
        b32 = floor((cap // 8) * 5)
        print(f" {v:2}  | {sz:3}x{sz:<3d} |  {chars:4d} ", end="")
        print(f"| {bys:5d}", end="")
        print(f"| {b32:7d}", end="")

        for sp in [2, 5, 10]:
            mx = b32 * sp
            print(f" | {mx:8d}", end="")

        print("")


# EOF



#
# random_sampler.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#


class RandomSampler:
    def __init__(self, probs):
        for p in probs:
            assert p > 0

        # Normalize given probabilities
        total = sum(probs)
        assert total > 0

        n = len(probs)

        P = []
        for p in probs:
            P.append((p * float(n)) / total)

        S = []
        L = []

        # Set separate index lists for small and large probabilities:
        for i in reversed(range(0, n)):
            # at variance from Schwarz, we reverse the index order
            if P[i] < 1:
                S.append(i)
            else:
                L.append(i)

        # Work through index lists
        _probs = [0] * n
        _aliases = [0] * n

        while len(S) > 0 and len(L) > 0:
            a = S.pop()  # Schwarz's l
            g = L.pop()  # Schwarz's g
            _probs[a] = P[a]
            _aliases[a] = g
            P[g] += P[a] - 1
            if P[g] < 1:
                S.append(g)
            else:
                L.append(g)

        while len(L) > 0:
            _probs[L.pop()] = 1

        while len(S) > 0:
            # can only happen through numeric instability
            _probs[S.pop()] = 1

        self.probs = _probs
        self.aliases = _aliases

    def next(self, rng_func):
        r1 = rng_func()
        r2 = rng_func()
        n = len(self.probs)
        i = int(float(n) * r1)
        return i if r2 < self.probs[i] else self.aliases[i]



#
# ur_encoder.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .bytewords import *
from .fountain_encoder import FountainEncoder


class UREncoder:
    # Start encoding a (possibly) multi-part UR.
    def __init__(self, ur, max_fragment_len, first_seq_num=0, min_fragment_len=10):
        self.ur = ur
        self.fountain_encoder = FountainEncoder(ur.cbor, max_fragment_len, first_seq_num, min_fragment_len)

    # Encode a single-part UR.
    @staticmethod
    def encode(ur):
        body = Bytewords.encode(Bytewords_Style_minimal, ur.cbor)
        return UREncoder.encode_ur([ur.type, body])

    def last_part_indexes(self):
        return self.fountain_encoder.last_part_indexes()

    # `True` if the minimal number of parts to transmit the message have been
    # generated. Parts generated when this is `true` will be fountain codes
    # containing various mixes of the part data.
    def is_complete(self):
        return self.fountain_encoder.is_complete()

    # `True` if this UR can be contained in a single part. If `True`, repeated
    # calls to `next_part()` will all return the same single-part UR.
    def is_single_part(self):
        return self.fountain_encoder.is_single_part()

    def next_part(self):
        part = self.fountain_encoder.next_part()
        if self.is_single_part():
            return UREncoder.encode(self.ur)
        else:
            return UREncoder.encode_part(self.ur.type, part)

    @staticmethod
    def encode_part(type, part):
        seq = "{}-{}".format(part.seq_num, part.seq_len)
        body = Bytewords.encode(Bytewords_Style_minimal, part.cbor())
        result = UREncoder.encode_ur([type, seq, body])
        return result

    @staticmethod
    def encode_uri(scheme, path_components):
        path = "/".join(path_components)
        return ":".join([scheme, path])

    @staticmethod
    def encode_ur(path_components):
        return UREncoder.encode_uri("ur", path_components)



import logging

logger = logging.getLogger(__name__)


import time
from typing import List

import mss
import numpy as np
import pygame
import pygame.camera
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtCore import QEvent, pyqtSignal


class BarcodeData:
    def __init__(self, data, rect):
        self.data = data  # The decoded text of the barcode
        self.rect = rect  # The bounding rectangle of the barcode as a tuple (x, y, width, height)

    def __repr__(self):
        return f"BarcodeData(data={self.data}, rect={self.rect})"


class VideoWidget(QtWidgets.QWidget):
    signal_qr_data_callback = pyqtSignal(object)

    def __init__(self, qr_data_callback=None, parent=None):
        super().__init__(parent)
        self.cv2 = None
        self.pyzbar = None
        try:
            # check if loading works
            # it depend on zlib installed in the os
            from pyzbar import pyzbar

            self.pyzbar = pyzbar
            logger.info("Load pyzbar successful")
        except:
            logger.info("Could not load pyzbar. Trying to load fallback cv2")
            import cv2

            self.cv2 = cv2

        self.is_screen_capture = False
        self.screen_capturer = mss.mss()

        self.label_image = QtWidgets.QLabel()
        self.qr_data_callback = qr_data_callback
        self.signal_qr_data_callback.connect(qr_data_callback)

        pygame.camera.init()
        self.cameras = self.get_valid_cameras()
        self.combo_cameras = QtWidgets.QComboBox()
        self.combo_cameras.addItems(self.cameras)
        self.combo_cameras.currentIndexChanged.connect(self.switch_camera)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.label_image)
        layout.addWidget(self.combo_cameras)

        self.setLayout(layout)

        self.capture = None

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_frame)
        self.timer.start(int(1000.0 / 30.0))  # 30 FPS

        # Add "Screen" option for screen capture
        self.combo_cameras.addItem("Screen")
        self.cameras.append("Screen")  # Adding "Screen" to the cameras list
        self.switch_camera(0)

    def closeEvent(self, event: QEvent) -> None:
        self.timer.stop()
        if self.capture:
            try:
                self.capture.stop()
            except:
                pass

        # If you call the parent's closeEvent(), it will proceed to close the widget
        super().closeEvent(event)

    def get_valid_cameras(self):
        valid_cameras = []
        for camera in pygame.camera.list_cameras():
            try:
                temp_camera = pygame.camera.Camera(camera, (640, 480))
                temp_camera.start()
                temp_camera.stop()
                valid_cameras.append(camera)
            except SystemError:
                continue
        return valid_cameras

    def switch_camera(self, index: int):
        selected_camera = self.combo_cameras.currentText()
        if selected_camera == "Screen":
            self.is_screen_capture = True
            # Additional logic for initializing screen capture can be added here
        else:
            self.is_screen_capture = False

            if self.capture:
                try:
                    self.capture.stop()
                except:
                    pass

            self.capture = pygame.camera.Camera(self.cameras[index], (640, 480))
            self.capture.start()
            time.sleep(0.1)  # Add a short delay

    def _numpy_to_surface(self, numpy_image: np.ndarray) -> pygame.Surface:
        # Convert numpy image (RGB) to pygame surface
        return pygame.surfarray.make_surface(numpy_image.transpose((1, 0, 2)))

    def _on_draw_surface(self, surface, barcode: BarcodeData):
        x, y, w, h = barcode.rect
        pygame.draw.rect(surface, (0, 255, 0), (x, y, w, h), 2)

    def get_barcodes(self, array: np.ndarray) -> List[BarcodeData]:
        if self.pyzbar:
            decoded_codes = self.pyzbar.decode(array)
            return [BarcodeData(data=decoded.data, rect=decoded.rect) for decoded in decoded_codes]
        elif self.cv2:
            array = self.cv2.transpose(array)
            array = self.cv2.cvtColor(array, self.cv2.COLOR_RGB2BGR)
            # Use OpenCV's QRCodeDetector to detect and decode the QR code.
            detector = self.cv2.QRCodeDetector()
            val, points, straight_qrcode = detector.detectAndDecode(array)

            barcodes = []
            if val:
                # If a QR code is detected, 'val' contains the decoded text.
                # 'points' contain the coordinates of the QR code corners.
                if points is not None:
                    # Convert points to a more manageable format
                    points = points[0]  # Points are returned in a nested array.

                    # Calculate the bounding rectangle using the points.
                    y, x, w, h = self.cv2.boundingRect(points.astype(np.int32))
                    rect = (x, y, w, h)  # Store the rectangle as a tuple

                    barcode = BarcodeData(val.encode(), rect)
                    barcodes.append(barcode)

            return barcodes
        else:
            return []

    def update_frame(self):

        barcodes = []
        if self.is_screen_capture:
            with mss.mss() as sct:
                monitor = sct.monitors[1]  # capture the first monitor
                sct_img = sct.grab(monitor)
                image = np.array(sct_img)

                # Convert BGRA to RGB by reordering the channels
                image = image[:, :, [2, 1, 0]]  # Rearrange BGR to RGB

                # Convert numpy image to surface for drawing
                surface = self._numpy_to_surface(image)
                surface = pygame.transform.flip(surface, True, False)
        else:
            try:
                surface = self.capture.get_image()
            except:
                # print("Could not get image")
                return

        barcodes = self.get_barcodes(pygame.surfarray.array3d(surface))
        surface = pygame.transform.flip(surface, False, True)
        surface = pygame.transform.rotate(surface, -90)

        for barcode in barcodes:
            self._on_draw_surface(surface, barcode)
            # only show the 1. barcode
            break

        surface = pygame.transform.flip(surface, False, True)
        self.showSurface(surface, scale_to=(640, 480))

        for barcode in barcodes:
            self.signal_qr_data_callback.emit(barcode.data)
            break

    def showSurface(self, surface: pygame.Surface, scale_to=(640, 480)):

        array3d = pygame.surfarray.array3d(surface)
        height, width, _ = array3d.shape
        bytes_per_line = 3 * width
        q_image = QtGui.QImage(array3d.data, width, height, bytes_per_line, QtGui.QImage.Format.Format_RGB888)

        # Convert QImage to QPixmap
        pixmap = QtGui.QPixmap.fromImage(q_image)

        # Scale the pixmap if scaling dimensions are provided
        if scale_to:
            pixmap = pixmap.scaled(
                scale_to[0],
                scale_to[1],
                QtCore.Qt.AspectRatioMode.KeepAspectRatio,
                QtCore.Qt.TransformationMode.SmoothTransformation,
            )

        self.label_image.setPixmap(pixmap)


class DemoVideoWidget(VideoWidget):
    def __init__(self, parent=None):
        super().__init__(qr_data_callback=self.show_qr, parent=parent)

        self.label_qr = QtWidgets.QTextEdit()

        self.layout().addWidget(self.label_qr)

    def show_qr(self, qr_data):
        s = qr_data.decode("utf-8")
        print(s)
        self.label_qr.setText(s)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from io import BytesIO

import segno
from PIL import Image

logger = logging.getLogger(__name__)


class QRGenerator:
    @staticmethod
    def create_qr_PILimage(data: str, scale=10, color="black", background="white") -> Image:
        """Generate a QR Code from the provided data and return it as an PIL Image

        :param data: The data to encode in the QR Code.
        :param scale: The scaling factor for the QR Code.
        :param color: Color of the QR code.
        :param background: Background color of the QR code (None for
            transparent).
        :return: QR Code as SVG string.
        """
        qr = segno.make(data)
        buffer = BytesIO()
        qr.save(buffer, kind="png", scale=scale, dark=color, light=background)  # Save QR code to buffer
        buffer.seek(0)  # Reset buffer's position to the beginning
        img = Image.open(buffer)  # Open the image using PIL
        return img

    @staticmethod
    def create_qr_svg(content: str, scale=1, color="black", background="white", encoding=None):
        """Generate a QR Code from the provided data and return it as an SVG
        string.

        :param content [str, int, bytes ]: The data to encode. Either a Unicode string, an integer or
                bytes. If bytes are provided, the `encoding` parameter should be
                used to specify the used encoding.
        :param encoding [str or None]: Indicates the encoding in mode "byte". By default
                (`encoding` is ``None``) the implementation tries to use the
                standard conform ISO/IEC 8859-1 encoding and if it does not fit, it
                will use UTF-8. Note that no ECI mode indicator is inserted by
                default (see :paramref:`eci <segno.make.eci>`).
                The `encoding` parameter is case insensitive.
        :param scale: The scaling factor for the QR Code.
        :param color: Color of the QR code.
        :param background: Background color of the QR code (None for transparent).
        :return: QR Code as SVG string.
        """
        try:
            qr = segno.make(content, encoding=encoding)
            svg = qr.svg_inline(scale=scale, dark=color, border=2, light=background)
        except Exception as e:
            logger.error(str(e))
            return
        return svg


if __name__ == "__main__":
    example_data = "Hello, QR Code!" * 50
    qr_image = QRGenerator.create_qr_PILimage(example_data)
    if qr_image:
        qr_image.show()  # Display the image
        qr_image.save("example_qr_code.png")  # Save the image to a file
    else:
        logger.error("Failed to create an example QR code.")



#!/usr/bin/env python
#
# (c) Copyright 2023 by Coinkite Inc. This file is in the public domain.
#
# This code will be added to you path when you do "pip install" on the BBQr package.
#
#
import io
import os
import random
import sys

import click
from bbqr import join_qrs, split_qrs
from bbqr.consts import FILETYPE_NAMES, KNOWN_FILETYPES


# Cleanup display (supress traceback) for user-feedback exceptions
# _sys_excepthook = sys.excepthook
def my_hook(ty, val, tb):
    if ty in {AssertionError, ValueError}:
        print("\n\n%s" % val, file=sys.stderr)
    else:
        return _sys_excepthook(ty, val, tb)


# sys.excepthook=my_hook


# Options we want for all commands
@click.group()
@click.option("--pdb", is_flag=True, help="Prepare patient for surgery to remove bugs.")
def main(**kws):
    # implement PDB option here
    if kws.pop("pdb", False):
        import pdb
        import sys

        def doit(ex_cls, ex, tb):
            pdb.pm()

        sys.excepthook = doit


@main.command("table")
def show_table():
    """Dump a table used in the spec"""
    from bbqr import tables

    tables.dump_table()


@main.command("decode")
@click.option("--raw", "-r", help="Output data as raw binary", is_flag=True)
def decode_bbqr(raw):
    """Undo a received BBQr series, back into useful data."""

    if sys.stdin.isatty():
        print(f"Paste data received, in any order here. Newlines between them.", file=sys.stderr)

    lines = [ln.strip() for ln in sys.stdin.readlines() if ln.strip()]

    try:
        file_type, data = join_qrs(lines)
    except Exception as exc:
        print(f"Error: {exc}")
        return 1

    if raw:
        click.get_binary_stream("stdout").write(data)
        return 0

    if file_type == "J":
        # pretty-print JSON
        import json

        j = json.loads(data)

        print(json.dumps(j, indent=2))

    elif file_type == "U":
        data = data.decode("utf-8")
        print(data)

    elif file_type == "P":
        from base64 import b64encode

        print("PSBT File:", file=sys.stderr)
        print(b64encode(data).decode("ascii"))
    elif file_type == "T":
        print("Bitcoin Transaction:", file=sys.stderr)
        from binascii import b2a_hex

        print(b2a_hex(data))

    elif file_type in "XB":
        print(f"{len(data)} bytes of Binary data... (not shown)", file=sys.stderr)

    elif file_type == "C":
        print(f"{len(data)} bytes of raw CBOR data... (not shown)", file=sys.stderr)

    else:
        print(f"Unknown file type code: {file_type}")


@main.command("make")
@click.argument("infile", type=click.File("rb"))
@click.option(
    "--encoding",
    "-e",
    metavar="(char)",
    default=None,
    type=click.Choice("H2Z"),
    help="Force low-level encoding: H 2 or Z",
)
@click.option(
    "--filetype",
    "-t",
    metavar="(char)",
    default=None,
    type=click.Choice(KNOWN_FILETYPES),
    help="Force specific file type code: " + "".join(KNOWN_FILETYPES),
)
@click.option(
    "--max-version",
    "-v",
    metavar="[1-40]",
    default=40,
    help="Max QR version to use (limits size, default unlimited: 40)",
)
@click.option(
    "--min-split", "-m", metavar="NUM", default=1, help="Produce at least this many QR codes (default: 1)"
)
@click.option(
    "--frame-delay",
    "-d",
    metavar="[ms/fr]",
    default=250,
    type=int,
    help="Delay between frame of animation (default: 250ms)",
)
@click.option(
    "--scale",
    "-s",
    metavar="NUM",
    default=4,
    help="For image outputs, the size of each QR pixel (default: 4)",
)
@click.option(
    "--outfile",
    "-o",
    metavar="filename.png",
    help="Name for output file",
    default=None,
    type=click.Path(dir_okay=False, writable=True, allow_dash=True),
)
@click.option("--fake-data", help="Generate huge empty data", type=int)
@click.option("--randomize-order", "-r", help="Shuffle output parts into random ordering", is_flag=True)
def make_qrs(
    randomize_order,
    infile=None,
    outfile=None,
    encoding=None,
    scale=4,
    max_version=40,
    frame_delay=250,
    min_split=1,
    fake_data=None,
    filetype=None,
):
    """Encode file as a series of QR codes"""

    if fake_data:
        # for Mk4/Q: maximum psbt size
        raw = bytes(fake_data)
    else:
        raw = infile.read()

    assert len(raw) > 5, "Input data too short?!"

    if not filetype:
        if ".psb" in infile.name.lower():
            filetype = "P"
            if raw[0:5] != b"psbt\xff":
                if raw[0:10].decode().isprintable():
                    print("Someone has saved Base64 or Hex encoded PSBT to disk? We want raw meat.")
                raise ValueError(infile.name)
        elif raw[0:8] in {b"01000000", b"02000000"}:
            # transaction in hex format
            filetype = "T"
            raw = bytes.fromhex(raw.decode("ascii"))
        elif raw[0:4] in {b"\x01\x00\x00\x00", b"\x02\x00\x00\x00"}:
            # binary transaction
            filetype = "T"
        elif raw[0] == b"{":
            # probably JSON
            filetype = "J"
        else:
            # otherwise text or binary
            try:
                raw.decode("utf-8")
                filetype = "U"
            except UnicodeError:
                filetype = "B"

        print(f"Detected file type: {filetype} -> {FILETYPE_NAMES[filetype]}", file=sys.stderr)

    vers, parts = split_qrs(
        raw, type_code=filetype, encoding=encoding, max_version=max_version, min_split=min_split
    )

    num_parts = len(parts)

    if len(parts) == 1:
        print(f"A single QR version {vers} will be needed.", file=sys.stderr)
    else:
        print(f"Need {num_parts} QR's each of version {vers}.", file=sys.stderr)

    if randomize_order:
        random.shuffle(parts)

    if not outfile or outfile == "-":
        for p in parts:
            print(p)
        return 0

    rootpath, ext = os.path.splitext(outfile)
    ext = ext.lower()[1:]

    if ext not in {"png", "svg", "gif"}:
        print(f"Unsupported output file type: {ext}")
        return 1

    # Render graphics -- very slow!
    import pyqrcode

    print("Building QR images... ", file=sys.stderr, end="", flush=True)
    qs = [pyqrcode.create(data, error="L", version=vers, mode="alphanumeric") for data in parts]
    print("done!", file=sys.stderr)

    if ext == "svg":
        # limitation: doesn't include progress bar animation
        for i in range(num_parts):
            fn = f"{rootpath}-{i+1}.{ext}" if num_parts > 1 else outfile
            qs[i].svg(open(fn, "wb"), scale=scale)
            print(f"Created file {fn!r}")

    elif ext in {"png", "gif"}:
        from PIL import Image, ImageChops, ImageDraw

        frames = []

        for i in range(num_parts):
            xbm = qs[i].xbm(scale=scale, quiet_zone=10)
            img = ImageChops.invert(Image.open(io.BytesIO(xbm.encode()))).convert("L")
            if num_parts > 1:
                # add progress bar
                pw = img.width // num_parts
                lm = (img.width - (pw * num_parts)) // 2
                draw = ImageDraw.Draw(img)
                h = scale // 2
                y = img.height - h - (scale // 2) - 1

                for j in range(num_parts):
                    draw.rectangle(
                        (lm + (j * pw), y, lm + ((j + 1) * pw), y + h), fill=(128 if i != j else 0)
                    )

            frames.append(img)

        if num_parts == 1:
            frames[0].save(outfile)
        else:
            frames[0].save(
                outfile,
                format=ext,
                save_all=True,
                loop=0,
                duration=frame_delay,
                default_image=False,
                append_images=frames[1:],
            )

        print(f"Created {outfile!r} with {len(frames)} frames.")


# EOF
